(***********************************************************************)
(*                                                                     *)
(*                       Ambients in JoCaml                            *)
(*                                                                     *)
(*               Alan Schmitt, projet Para, INRIA Rocquencourt         *)
(*                                                                     *)
(*  Copyright 1999 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

open Syntax
open Print

type name = string

type sub_in_t = << Id.t * string * <<unit>> * int >>
and sub_out_t = << Id.t * string * <<unit>> * int >>
    
and live_t = << subambient_interface * Id.t * ambient_interface >>
and a_in_t = << name * << unit >> >>
and a_out_t = << name * << unit >> >>
and a_open_t = << name * << unit >> >>
and comm_in_t = << << capacity >> >>
and comm_out_t = << capacity * << unit >> >>
and commit_t = << subambient_interface * << unit >> * int  >>
and opening_t = << ambient_interface >>
and cont_t = << << subambient_interface * Id.t * ambient_interface >> * subambient_interface * Id.t * ambient_interface >>
and sub_t = << Id.t * subcontent_t * cont_t  >>
    
and subambient_interface = {
    there: Join.location; 
    sub: sub_t; 
    sub_in: sub_in_t; 
    sub_out: sub_out_t}

and ambient_interface = {
    e: subambient_interface; 
    a_in: a_in_t; 
    a_out: a_out_t; 
    a_open: a_open_t; 
    comm_in: comm_in_t; 
    comm_out: comm_out_t}
      
and subcontent_t = {
    ee: subambient_interface; 
    name: name; 
    commit: commit_t; 
    opening: opening_t; 
    tbang: <<unit>>; 
    to_string: unit -> string }
      
exception Amb_has_moved
exception Action_impossible
exception Amb_syntax_error
exception Bad_log_entry
exception Weird

(* List.filter not in jocaml $%^^?! *)

let rec list_filter p = function
  | a::l when p a -> a::list_filter p l
  | _::l -> list_filter p l
  | [] -> [] 

let clear_get lr = let l = !lr in lr := []; l

(* TRACE, now several trace functions, easily disabled *)

let ps s = print_string s (* stdout closure without arg *)

let yell s = ps ("``"^s^"''"); print_newline (); flush stdout
let translating p =  ps ("  translating "); print_process p; ps "\n"; flush stdout 
let translated p  = ps ("  forked "); print_process p; ps "\n"; flush stdout
let attempting s = ps ("  action "^s^"\n"); flush stdout
let forwarding n s = ps ("  forward "^s^" out of shadow ambient "^n^"\n"); flush stdout
let string_of_amb n id = (n^" "^Id.to_string id)
let print_id id = ps (Id.to_string id)

(* LOG MANAGEMENT *)
    
(* the following type is used for storing messages sent
   (as in the log) and for storing active action messages *)

type log_type = 
    IN of (Id.t * string * <<unit>> * int)
  | OUT of (Id.t * string * <<unit>> * int)
  | OPEN of (string * <<unit>>)

let action_to_string = function
  | IN (id, a, k, v) -> ("in "^a^" by ambient "^Id.to_string id)
  | OUT (id, a, k, v) -> ("out "^a^" by ambient "^Id.to_string id)
  | OPEN (a, k) -> ("open "^a) ;;

let print_active_messages n id l =  (* print active messages waiting to be used*)
  begin 
    ps ("  log for ambient "^string_of_amb n id^"  { ");
    let details s = ps (s^", ") in
    List.iter (function a -> ps (action_to_string a^", ")) l;
    ps " } .\n";
    flush stdout
  end

let log_add v lr = lr := v::!lr

let log_remove cont lr =
  let rec flt = function 
    | [] -> raise Weird
    | IN  (_, _, _, v)::l' when v = cont -> l' 
    | OUT (_, _, _, v)::l' when v = cont -> l'
    | a::l' -> a:: flt l' in
  lr := flt !lr

let log_flush a_in a_out lr = 
  List.iter (function
    | IN (_, a, k, _) -> spawn {a_in (a, k)} (* ps "in spawned\n"; flush stdout; *) 
    | OUT (_, a, k, _) -> spawn {a_out (a, k)}
    | _ -> raise Bad_log_entry (* no other logged messages *) ) 
    (clear_get lr)

let has_new_subambient n id n' id' l =
  ps ("  ambient "^string_of_amb n id^" has new subambient "^string_of_amb n' id'^"\n");
  print_active_messages n id l

let make_new_id = (* creates a fresh ambient name *)
  let c = ref 0 in 
  function n -> (incr c; (Name (ResolvedName ("nu"^string_of_int n^","^string_of_int !c))))

let sub_name_to_capa env = function (* substitutes a name into a capacity *)
    Ident s -> (
      try List.assoc s env
      with Not_found -> Name (ResolvedName s))
  | ResolvedName s -> Name (ResolvedName s)

let sub_name_to_name env = function (* substitutes a name into a name *)
    Ident s -> (
      try 
	match (List.assoc s env) with
	  Name (ResolvedName n) -> ResolvedName n
	| _ as c -> (ps "syntax error: "; 
		     print_capa c; 
		     ps "instead of name"; 
		     flush stdout; 
		     raise Amb_syntax_error)
      with Not_found -> ResolvedName s)
  | ResolvedName s -> ResolvedName s

let rec substitute env = function (* substitutes in a capacity the values associated to the identifiers *)
  | In n -> In (sub_name_to_name env n)
  | Out n -> Out (sub_name_to_name env n)
  | Open n -> Open (sub_name_to_name env n)
  | Name n -> sub_name_to_capa env n
  | Yell s -> Yell s
  | Seq_capa (c1, c2) -> Seq_capa ((substitute env c1), (substitute env c2))
   
let rec trad f env bang n p = (* AST interpreter *)
  let guard p = ( 
    match bang with
    | None -> let def k! () = {trad f env None n p ;} in k
    | Some bp -> let def k! () = {trad f env None n p ;} | bp () in k ) in
  let fetch n = (* never return a capacity containing an Ident *)
    try List.assoc n env
    with Not_found -> Name (ResolvedName n) in
  let resolve n = 
    match n with
      Ident i -> ( 
	match fetch i with
	| Name (ResolvedName r) -> r
	| _ -> raise Amb_syntax_error )
    | ResolvedName r -> r
  in

  match p with
    Par (p, q) -> spawn { {trad f env bang n p ;} | {trad f env bang n q ;} }
  | Seq (Seq_capa (c1, c2), p)  -> trad f env bang n (Seq (c1, (Seq (c2, p))))
  | Seq (In name, p) -> 
      let name' = resolve name in translated (Seq(In (ResolvedName name'),p));
      spawn {f.a_in(name', guard p)} 
  | Seq (Out name, p) -> 
      let name' = resolve name in translated (Seq(Out (ResolvedName name'),p));
      spawn {f.a_out(name', guard p)}
  | Seq (Open name, p) -> 
      let name' = resolve name in translated (Seq(Open (ResolvedName name'),p));
      spawn {f.a_open(name', guard p)}
  | Seq (Name name, p) -> begin
      match name with
	ResolvedName str -> raise Amb_syntax_error
      |	Ident str -> begin
	  match fetch str with
	  | Name _ -> raise Amb_syntax_error
	  | Seq_capa (c1, c2) -> trad f env bang n (Seq (c1, Seq (c2, p)))
	  | other -> trad f env bang n (Seq (other, p)) end
  end
  | Seq (Yell str, p) -> yell str ; spawn {trad f env bang n p; }
  | Amb (name, p, b) -> 
      let name' = resolve name in
      translating (Amb((ResolvedName name'),p,b));
      ambient name' f.e env p b bang n
  | New (name, p) ->
      trad f ( (name, (make_new_id n)) :: env ) bang n p
  | Comm_in (avar, p) as witness -> begin
      let guard = match bang with
      | None ->  let def k! value = {trad f ((avar, value) :: env) None n p ;} in k
      | Some bp -> let def k! value = {trad f ((avar, value) :: env) None n p ;} | bp () in k in
      translated witness;
      spawn {f.comm_in guard}  end
  | Comm_out c as witness-> begin
      let val_to_send =	(substitute env c) in
      let guard = ( match bang with
      | None -> let def k! () = {} in k
      | Some bp -> let def k! () = bp () in k ) in
      translated witness;
      spawn {f.comm_out (val_to_send, guard)}  end
  | Bang p ->
      let def bp! () = {trad f env (Some bp) n p;} in
      spawn { bp () | bp () }
  | Zero -> ()
	
and ambient name initial_e env code_to_run reg bang base_id =
  let loc here
  do {
    let subs = Sub.create () in
    let local_log = ref [] in
    let active_messages = ref [] in
    let incr_msg_index = 
      let r = ref 0 in 
      function () -> (let v = !r in incr r; v) in
    let isbang = ref bang in
    let attempt action e id' f test_bang =
      begin
        attempting (action_to_string action^" within "^string_of_amb name id');
        if (not (use_message subs name f e test_bang action)) then
          (log_add action active_messages;
	   attempting (action_to_string action^" deferred"))
      end in

    (* --- AMBIENT SYNCHRONIZATION --- *)

    (* sub stored in hashtables *)
    let def live! (e', id', f) | sub! (id, s, cont) =
      Sub.add subs s.name id s;
      has_new_subambient name id' s.name id !active_messages;
      active_messages := list_filter (* search for new actions, discarding others *)
           (function msg -> not (use_message subs name f e' test_bang msg)) 
           !active_messages;
      { cont (live, e', id', f) }
	
    (* subjective in/out, logged and forwarded up *)
    or live! (e, id, f) | a_in! (a, k) =
      let v = incr_msg_index () in 
      ps ("  in "^a^" by "^string_of_amb name id^" logged with index "^string_of_int v^"\n");flush stdout;  
      log_add (IN(id, a, k, v)) local_log;
      live (e, id, f) | e.sub_in(id, a, k, v) 
	
    or live! (e, id, f) | a_out! (a, k) =
      let v = incr_msg_index () in
      log_add (OUT(id, a, k, v)) local_log;
      live (e, id, f) | e.sub_out(id, a, k, v)
	
    (* sub_in/sub_out/open, attempted, otherwise stored in hashtables *)
    or live! (e, id', f) | sub_in! (id, a, k, v) = 
      attempt (IN (id, a, k, v)) e id' f test_bang; live(e, id', f)

    or live! (e, id', f) | sub_out! (id, a, k, v) = 
      attempt (OUT(id, a, k, v)) e id' f test_bang; live(e,id',f)

    or live! (e, id', f) | a_open! (a, k) = 
      attempt (OPEN(a, k)) e id' f test_bang; live(e, id', f)
	   
    (* communication; the hidden continuation is used for bangs *)
    or live!(e, id, f) | comm_in! k | comm_out! (v, k') = live(e, id, f) | k(v) | k' ()
       
    (* migration; the log is flushed toward the new parents *)
    or live! (e, id, f) | commit! ((e':subambient_interface), k, v) =
      go(e'.there);
      (try log_remove v local_log
      with Weird -> ps ("\n weird log_remove, ambient "^Id.to_string id^", log entry "^string_of_int v^"\n\n"));
        let id' = Id.next id in
      let new_e = {there=here; sub=sub; sub_in=sub_in; sub_out=sub_out} in
       (* This continuation makes sure that the sub message has been
	  recorded in the hash table before the log is flushed.
	  The live message is received as an argument to make sure
	  no message will be added to the log during the flush.
	  This will be replaced by a synchronous call to sub *)
       let def cont!(l, e, i, f) = {
         log_flush a_in a_out local_log;
         spawn {l(e, i, f) | k ()};
       } in
       let s = {ee= new_e; name=name; commit=commit; opening=opening; tbang=test_bang; to_string=to_string} in
       spawn {live (e', id', f) | e'.sub(id',s, (cont:cont_t))};
      
    (* dissolution *) 
    or live! (e, id, f) | opening! (parent) =
       (* go (parent.e.here); *)
       {  opened(parent)
       | { log_flush parent.a_in parent.a_out local_log;} (* the messages in and out are emitted in the new local ambient *)
       | { flush_actions parent.e.sub_in parent.e.sub_out parent.a_open active_messages;} (* the active messages are also flushed directly in the new local ambient *)
       | { let def foo! (l, e, i, f) = {} in  (* dummy continuation *)
           Sub.flush subs (* local subs are flushed directly in new local ambient*)
           (fun n id v -> spawn {parent.e.sub(id, v, foo)});  }}  

    (* once opened, all requests get forwarded upward *)
    or opened!(up) | a_in!(a, k)     = opened (up) | up.a_in(a, k)
    or opened!(up) | a_out!(a, k)    = opened (up) | up.a_out(a, k)
    or opened!(up) | a_open!(a, k)   = opened (up) | up.a_open(a, k)
    or opened!(up) | comm_in!(k)     = opened (up) | up.comm_in(k)
    or opened!(up) | comm_out!(a, k) = opened (up) | up.comm_out(a, k)
    or opened!(up) |     sub! (id,s,c) = 
       opened (up) | up.e.sub (id,s,c)
     | { forwarding name ("sub message of "^string_of_amb s.name id); } 
    or opened!(up) | sub_in!(id, a, k, v) =  opened (up) | up.e.sub_in(id, a, k, v) 
     | { forwarding name ("in "^a^" message"); } 
    or opened!(up) | sub_out!(id, a, k, v) = opened (up) | up.e.sub_out(id, a, k, v)
     | { forwarding name ("out "^a^" message"); } 
      
    (* This channel deals with replication*)
    or live! (e, id, f) | test_bang! () =
       ( match (!isbang) with
         | None -> ()
         | Some bp -> isbang := None; spawn { bp () });
       live (e, id, f)
       
    
    or opened!(up) | test_bang! () =
       ( match (!isbang) with
         | None -> ()
         | Some bp -> isbang := None; spawn { bp () });
       opened(up)	   

    or live! (e, id, f) | to_string (c) =
      let snapshot = Sub.to_string subs (
        fun a id v -> a^v.to_string () ) in
      live (e, id, f) | reply (snapshot)
     
    in {
      let id = Id.make base_id in
      let e = {there = here; sub = sub; sub_in = sub_in; sub_out = sub_out} in
      let f = {e = e; a_in = a_in; a_out = a_out; a_open = a_open; comm_in = comm_in; comm_out = comm_out } in
      let def foo! (l, e, i, f') = l(e, i, f') 
        | {trad f env 
	    (match (!isbang) with
	    | None -> None
	    | Some bp -> Some test_bang)
	 base_id code_to_run ;}
      in
      spawn {initial_e.sub(id, {ee=e; name=name; commit=commit; opening=opening; tbang=test_bang; to_string=to_string}, foo)};
      ps ("  new ambient "^string_of_amb name id^" [ "); print_process code_to_run; ps " ]\n"; flush stdout; 
      (if reg then (
        Ns.register name f vartype;
        ps ("  "^name^" has been registered for remote access\n"); flush stdout ));
      {live (initial_e, id, f) }}} (* end location *)
  in ()
      
(* the following functions resend the active messages that are in the list of active messages *)
and flush_actions sub_in sub_out a_open lr = 
  List.iter (function
    | IN (id, a, k, v) -> spawn {sub_in (id, a, k, v)}
    | OUT (id, a, k, v) -> spawn {sub_out (id, a, k, v)}
    | OPEN (a, k) -> spawn {a_open (a, k)} ) (clear_get lr)
	   
and use_message subs name f e test_bang action = 
  try (match action with 
  | IN (id, a, k, v) -> (
      let amb1 = 
        try Sub.find_id subs id (* check that the moving ambient is still present *) 
        with Not_found -> raise Amb_has_moved in
      let id2,amb2 = 
        try Sub.find_name_different subs a id
        with Not_found -> raise Action_impossible in
      ps ("  ambient "^string_of_amb amb1.name id^" moves in "^string_of_amb amb2.name id2^"\n"); flush stdout;
      Sub.remove subs id; (* remove sub entry for the moving ambient *)
      spawn {amb1.commit(amb2.ee, k, v) | amb1.tbang () | amb2.tbang ()}; ) 
  | OUT (id, a, k, v) -> (
      let amb1 = 
        try Sub.find_id subs id (* check that the moving ambient is still present *) 
        with Not_found -> raise Amb_has_moved in
      if not (a = name) then raise Action_impossible;
      ps ("  ambient "^string_of_amb amb1.name id^" moves out of "^name^"\n"); flush stdout;
      Sub.remove subs id;
      spawn {amb1.commit(e, k, v) | amb1.tbang () | test_bang ()};)
  | OPEN (a, k) -> (    
      let id_a, amb_a = 
        try Sub.find_name subs a 
        with Not_found -> raise Action_impossible in
      ps ("  ambient "^string_of_amb amb_a.name id_a^" is being opened\n"); flush stdout; 
      Sub.remove subs id_a; (* removes sub entry for the opened ambient *)
      spawn {amb_a.opening (f) | k () | amb_a.tbang ()};));
    
    true (* true is the action is used-up, either done or discarded *)
  with
  | Action_impossible -> false
  | Amb_has_moved -> ps ("  stale "^action_to_string action^" discarded\n");flush stdout; true
